import functools
import typing

import parsec
import typing_extensions

typing_extensions.TypeAlias

from .dynamic import (
    TalonCapture,
    TalonChoice,
    TalonCommandDeclaration,
    TalonComment,
    TalonDeclarations,
    TalonEndAnchor,
    TalonList,
    TalonOptional,
    TalonParenthesizedRule,
    TalonRepeat,
    TalonRepeat1,
    TalonRule,
    TalonSeq,
    TalonSourceFile,
    TalonStartAnchor,
    TalonWord,
)

################################################################################
# Export typing.List
################################################################################

__all__: typing.List[str] = [
    "AnyTalonRule",
    "ListValue",
    "find_command",
    "match",
]


################################################################################
# Definitions
################################################################################

AnyTalonRule: typing_extensions.TypeAlias = typing.Union[
    TalonCapture,
    TalonChoice,
    TalonEndAnchor,
    TalonList,
    TalonOptional,
    TalonParenthesizedRule,
    TalonRepeat,
    TalonRepeat1,
    TalonRule,
    TalonSeq,
    TalonStartAnchor,
    TalonWord,
]

ListValue = typing.Union[typing.List[str], typing.Dict[str, typing.Any]]


################################################################################
# Implement 'match' and 'find_command'
################################################################################


def find_command(
    self: TalonSourceFile,
    text: typing.Sequence[str],
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[
        typing.Callable[[str], typing.Optional[ListValue]]
    ] = None,
) -> typing.Optional[TalonCommandDeclaration]:
    parser = _TalonSourceFile_to_TalonCommandDeclarationMatcher(
        self, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
    )
    try:
        return parser.parse_strict(text)
    except parsec.ParseError:
        return None


setattr(TalonSourceFile, "find_command", find_command)


def match(
    self: typing.Union[TalonCommandDeclaration, AnyTalonRule],
    text: typing.Sequence[str],
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[
        typing.Callable[[str], typing.Optional[ListValue]]
    ] = None,
) -> bool:
    try:
        matcher = _to_SequenceMatcher(
            self, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
        matcher.parse_strict(text)
        return True
    except parsec.ParseError:
        return False


setattr(TalonCommandDeclaration, "match", match)
setattr(TalonCapture, "match", match)
setattr(TalonChoice, "match", match)
setattr(TalonEndAnchor, "match", match)
setattr(TalonList, "match", match)
setattr(TalonOptional, "match", match)
setattr(TalonParenthesizedRule, "match", match)
setattr(TalonRepeat, "match", match)
setattr(TalonRepeat1, "match", match)
setattr(TalonRule, "match", match)
setattr(TalonSeq, "match", match)
setattr(TalonStartAnchor, "match", match)
setattr(TalonWord, "match", match)

################################################################################
# Generic '_to_SequenceMatcher'
################################################################################

# NOTE: The parsec library defines the types parsec.Parser and parsec.Value to
#       be generic _in the interface files_, but the runtime types don't extend
#       the Generic class, so we only supply the arguments when type checking.
if typing.TYPE_CHECKING:
    SequenceMatcher: typing_extensions.TypeAlias = parsec.Parser[
        typing.Optional[typing.NoReturn]
    ]
    SequenceMatcherValue: typing_extensions.TypeAlias = parsec.Value[
        typing.Optional[typing.NoReturn]
    ]
    TalonCommandDeclarationMatcher: typing_extensions.TypeAlias = parsec.Parser[
        TalonCommandDeclaration
    ]
    TalonCommandDeclarationMatcherValue: typing_extensions.TypeAlias = parsec.Value[
        TalonCommandDeclaration
    ]
else:
    SequenceMatcher: typing_extensions.TypeAlias = parsec.Parser
    SequenceMatcherValue: typing_extensions.TypeAlias = parsec.Value
    TalonCommandDeclarationMatcher: typing_extensions.TypeAlias = parsec.Parser
    TalonCommandDeclarationMatcherValue: typing_extensions.TypeAlias = parsec.Value


def _to_SequenceMatcher(
    node: typing.Union[TalonCommandDeclaration, AnyTalonRule],
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[
        typing.Callable[[str], typing.Optional[ListValue]]
    ] = None,
) -> SequenceMatcher:
    if isinstance(node, TalonCapture):
        return _TalonCapture_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonCommandDeclaration):
        return _TalonCommandDeclaration_to_TalonCommandDeclarationMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        ).result(None)
    elif isinstance(node, TalonChoice):
        return _TalonChoice_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonEndAnchor):
        return _TalonEndAnchor_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonList):
        return _TalonList_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonOptional):
        return _TalonOptional_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonParenthesizedRule):
        return _TalonParenthesizedRule_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonRepeat):
        return _TalonRepeat_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonRepeat1):
        return _TalonRepeat1_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonRule):
        return _TalonRule_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonSeq):
        return _TalonSeq_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonStartAnchor):
        return _TalonStartAnchor_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )
    elif isinstance(node, TalonWord):
        return _TalonWord_to_SequenceMatcher(
            node, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
        )


################################################################################
# Compile TalonSourceFile to Parser
################################################################################


def _TalonSourceFile_to_TalonCommandDeclarationMatcher(
    self: TalonSourceFile,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> TalonCommandDeclarationMatcher:
    buffer: typing.List[TalonCommandDeclarationMatcher] = []
    for declarations in self.children:
        if isinstance(declarations, TalonDeclarations):
            buffer.append(
                _TalonDeclarations_to_TalonCommandDeclarationMatcher(
                    declarations,
                    fullmatch=fullmatch,
                    get_capture=get_capture,
                    get_list=get_list,
                )
            )
    return functools.reduce(
        parsec.try_choice, buffer, parsec.fail_with(f"input does not match any command")
    )


################################################################################
# Compile TalonDeclarations to Parser
################################################################################


def _TalonDeclarations_to_TalonCommandDeclarationMatcher(
    self: TalonDeclarations,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> TalonCommandDeclarationMatcher:
    buffer: typing.List[TalonCommandDeclarationMatcher] = []
    for declaration in self.children:
        if isinstance(declaration, TalonCommandDeclaration):
            buffer.append(
                _TalonCommandDeclaration_to_TalonCommandDeclarationMatcher(
                    declaration,
                    fullmatch=fullmatch,
                    get_capture=get_capture,
                    get_list=get_list,
                )
            )
    return functools.reduce(
        parsec.try_choice, buffer, parsec.fail_with(f"input does not match any command")
    )


################################################################################
# Compile TalonCommandDeclaration to Parser
################################################################################


def _TalonCommandDeclaration_to_TalonCommandDeclarationMatcher(
    self: TalonCommandDeclaration,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> TalonCommandDeclarationMatcher:
    return _to_SequenceMatcher(
        self.rule, fullmatch=fullmatch, get_capture=get_capture, get_list=get_list
    ).result(self)


################################################################################
# Implement AnyTalonRule._to_SequenceMatcher
################################################################################


def _always() -> SequenceMatcher:
    def _always_parser(text: typing.Sequence[str], index: int) -> SequenceMatcherValue:
        return parsec.Value.success(index, None)

    return parsec.Parser(fn=_always_parser)


def _word(token: str) -> SequenceMatcher:
    def _word_parser(text: typing.Sequence[str], index: int) -> SequenceMatcherValue:
        try:
            if text[index] == token:
                return parsec.Value.success(index + 1, None)
        except IndexError:
            pass
        # NOTE: `Value.failure` creates a value of type `Value[<nothing>]`
        return typing.cast(
            SequenceMatcherValue, parsec.Value.failure(index, str(token))
        )

    return parsec.Parser(fn=_word_parser)


def _TalonCapture_to_SequenceMatcher(
    self: TalonCapture,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    capture_name = self.capture_name.text
    if get_capture:
        capture = get_capture(capture_name)
        if capture:
            return _to_SequenceMatcher(
                capture, get_capture=get_capture, get_list=get_list
            )
    return _word(f"<{capture_name}>").result(None)


def _TalonChoice_to_SequenceMatcher(
    self: TalonChoice,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    acc: typing.Optional[SequenceMatcher] = None
    for rule in reversed(self.children):
        if not isinstance(rule, TalonComment):
            parser = _to_SequenceMatcher(
                rule,
                fullmatch=fullmatch,
                get_capture=get_capture,
                get_list=get_list,
            )
            acc = parser ^ acc if acc else parser
    assert acc is not None
    return acc.result(None)


def _TalonEndAnchor_to_SequenceMatcher(
    self: TalonEndAnchor,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    if fullmatch:
        return parsec.eof()
    else:
        return _always()


def _TalonList_to_SequenceMatcher(
    self: TalonList,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    list_name = self.list_name.text
    if get_list:
        list_value = get_list(list_name)
        if isinstance(list_value, dict):
            list_value = list(list_value.keys())
        if isinstance(list_value, list):
            return parsec.one_of([item for item in list_value]).result(None)
    return _word(f"{{{list_name}}}").result(None)


def _TalonOptional_to_SequenceMatcher(
    self: TalonOptional,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    return parsec.optional(
        _to_SequenceMatcher(
            self.get_child(),
            fullmatch=fullmatch,
            get_capture=get_capture,
            get_list=get_list,
        )
    ).result(None)


def _TalonParenthesizedRule_to_SequenceMatcher(
    self: TalonParenthesizedRule,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    return _to_SequenceMatcher(
        self.get_child(),
        fullmatch=fullmatch,
        get_capture=get_capture,
        get_list=get_list,
    ).result(None)


def _TalonRepeat_to_SequenceMatcher(
    self: TalonRepeat,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    return parsec.many(
        _to_SequenceMatcher(
            self.get_child(),
            fullmatch=fullmatch,
            get_capture=get_capture,
            get_list=get_list,
        )
    ).result(None)


def _TalonRepeat1_to_SequenceMatcher(
    self: TalonRepeat1,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    return parsec.many1(
        _to_SequenceMatcher(
            self.get_child(),
            fullmatch=fullmatch,
            get_capture=get_capture,
            get_list=get_list,
        )
    ).result(None)


def _TalonRule_to_SequenceMatcher(
    self: TalonRule,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    acc: typing.Optional[SequenceMatcher] = None
    for rule in reversed(self.children):
        if not isinstance(rule, TalonComment):
            parser = _to_SequenceMatcher(
                rule,
                fullmatch=fullmatch,
                get_capture=get_capture,
                get_list=get_list,
            )
            acc = (parser >> acc) if acc else parser
    assert acc is not None
    return acc.result(None)


def _TalonSeq_to_SequenceMatcher(
    self: TalonSeq,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    acc: typing.Optional[SequenceMatcher] = None
    for rule in reversed(self.children):
        if not isinstance(rule, TalonComment):
            parser = _to_SequenceMatcher(
                rule,
                fullmatch=fullmatch,
                get_capture=get_capture,
                get_list=get_list,
            )
            acc = parser > acc if acc else parser
    assert acc is not None
    return acc.result(None)


def _TalonStartAnchor_to_SequenceMatcher(
    self: TalonStartAnchor,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    if fullmatch:

        def start_anchor_parser(
            text: typing.Sequence[str], index: int
        ) -> SequenceMatcherValue:
            if index == 0:
                return parsec.Value.success(index, None)
            else:
                return typing.cast(
                    SequenceMatcherValue,
                    parsec.Value.failure(index, "should be start of the input"),
                )

        return SequenceMatcher(fn=start_anchor_parser)
    else:
        return _always()


def _TalonWord_to_SequenceMatcher(
    self: TalonWord,
    *,
    fullmatch: bool = False,
    get_capture: typing.Optional[
        typing.Callable[[str], typing.Optional[AnyTalonRule]]
    ] = None,
    get_list: typing.Optional[typing.Callable[[str], typing.Optional[ListValue]]],
) -> SequenceMatcher:
    return _word(self.text).result(None)



from typing import List

from pytest import mark
from pytest_golden.plugin import GoldenTestFixture

from . import class_signatures, function_signatures, pyver


@mark.golden_test(f"data/api.{pyver()}.yml")  # type: ignore[misc]
def test_talon_api(golden: GoldenTestFixture) -> None:
    assert golden["input"] is None

    import tree_sitter_talon

    defns = {
        name: defn
        for name, defn in tree_sitter_talon.__dict__.items()
        if name in tree_sitter_talon.__all__
    }
    globals().update(defns)

    output: List[str] = []
    output.extend(function_signatures(defns))
    output.extend(class_signatures(defns))

    assert "\n".join(output) == golden.out["output"]



import json
import pathlib

import importlib_resources
import packaging.version

################################################################################
# Version Number
################################################################################

# The tree-sitter-talon version.
try:
    TREE_SITTER_TALON_VERSION: packaging.version.Version = packaging.version.Version(
        json.loads(
            importlib_resources.files("tree_sitter_talon")
            .joinpath("_tree_sitter_talon", "package.json")
            .read_text(encoding="utf-8")
        )["version"]
    )
except ModuleNotFoundError:
    TREE_SITTER_TALON_VERSION: packaging.version.Version = packaging.version.Version(  # type: ignore[no-redef]
        json.loads(
            (
                pathlib.Path(__file__).parent / "_tree_sitter_talon" / "package.json"
            ).read_text(encoding="utf-8")
        )["version"]
    )

VERSION: str = f"{TREE_SITTER_TALON_VERSION.major}!1.6"



import io
import pickle

from pytest import mark
from pytest_golden.plugin import GoldenTestFixture, GoldenTestUsageWarning

from tree_sitter_talon import (  # type: ignore[attr-defined]
    TalonCommandDeclaration,
    TalonSourceFile,
    parse,
)

from . import node_dict_simplify


@mark.golden_test("data/golden/*.yml")  # type: ignore[misc]
def test_golden_dict(golden: GoldenTestFixture) -> None:
    try:
        node = parse(golden["input"])
        node_dict = node.to_dict()
        node_dict_simplify(node_dict)
        assert node_dict == golden.out["output"]
    except GoldenTestUsageWarning:
        pass


@mark.golden_test("data/golden/*.yml")  # type: ignore[misc]
def test_golden_pickle(golden: GoldenTestFixture) -> None:
    try:
        node_before = parse(golden["input"])
        node_pickled = pickle.dumps(node_before)
        node_after = pickle.loads(node_pickled)
        assert node_before == node_after
    except GoldenTestUsageWarning:
        pass


@mark.golden_test("data/golden/*.yml")  # type: ignore[misc]
def test_golden_match(golden: GoldenTestFixture) -> None:
    source_file = parse(golden["input"])
    assert isinstance(source_file, TalonSourceFile)
    try:
        for command in golden["commands"]:
            assert isinstance(command, str)
            declaration = source_file.find_command(list(command.split()))
            assert isinstance(declaration, TalonCommandDeclaration)
    except KeyError:
        pass
    except GoldenTestUsageWarning:
        pass



import warnings
from dataclasses import fields, is_dataclass
from functools import partial
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Type

from ._compat import _removeprefix
from ._internal.dynamic import (
    Branch,
    Leaf,
    Node,
    NodeFieldName,
    NodeTypeError,
    NodeTypeName,
    ParseError,
    Point,
    TalonAction,
    TalonArgumentList,
    TalonAssignmentStatement,
    TalonBinaryOperator,
    TalonBlock,
    TalonCapture,
    TalonChoice,
    TalonCommandDeclaration,
    TalonComment,
    TalonDeclaration,
    TalonDeclarations,
    TalonEndAnchor,
    TalonError,
    TalonExpression,
    TalonExpressionStatement,
    TalonFloat,
    TalonIdentifier,
    TalonImplicitString,
    TalonInteger,
    TalonInterpolation,
    TalonKeyAction,
    TalonKeyBindingDeclaration,
    TalonList,
    TalonMatch,
    TalonMatches,
    TalonMatchModifier,
    TalonNumber,
    TalonOperator,
    TalonOptional,
    TalonParenthesizedExpression,
    TalonParenthesizedRule,
    TalonRepeat,
    TalonRepeat1,
    TalonRule,
    TalonSeq,
    TalonSettingsDeclaration,
    TalonSleepAction,
    TalonSourceFile,
    TalonStartAnchor,
    TalonStatement,
    TalonString,
    TalonStringContent,
    TalonStringEscapeSequence,
    TalonTagImportDeclaration,
    TalonUnaryOperator,
    TalonVariable,
    TalonWord,
    from_tree_sitter,
    language,
    parse,
    parse_file,
    parser,
)
from ._internal.match import AnyTalonRule, ListValue, find_command, match
from ._version import VERSION

################################################################################
# Export List
################################################################################

__all__ = [
    "Branch",
    "Leaf",
    "Node",
    "NodeFieldName",
    "NodeTypeError",
    "NodeTypeName",
    "ParseError",
    "Point",
    "TalonAction",
    "TalonArgumentList",
    "TalonAssignmentStatement",
    "TalonBinaryOperator",
    "TalonBlock",
    "TalonCapture",
    "TalonChoice",
    "TalonCommandDeclaration",
    "TalonComment",
    "TalonDeclaration",
    "TalonDeclarations",
    "TalonEndAnchor",
    "TalonError",
    "TalonExpression",
    "TalonExpressionStatement",
    "TalonFloat",
    "TalonIdentifier",
    "TalonImplicitString",
    "TalonInteger",
    "TalonInterpolation",
    "TalonKeyAction",
    "TalonKeyBindingDeclaration",
    "TalonList",
    "TalonMatch",
    "TalonMatches",
    "TalonMatchModifier",
    "TalonNumber",
    "TalonOperator",
    "TalonOptional",
    "TalonParenthesizedExpression",
    "TalonParenthesizedRule",
    "TalonRepeat",
    "TalonRepeat1",
    "TalonRule",
    "TalonSeq",
    "TalonSettingsDeclaration",
    "TalonSleepAction",
    "TalonSourceFile",
    "TalonStartAnchor",
    "TalonStatement",
    "TalonString",
    "TalonStringContent",
    "TalonStringEscapeSequence",
    "TalonTagImportDeclaration",
    "TalonUnaryOperator",
    "TalonVariable",
    "TalonWord",
    "from_tree_sitter",
    "language",
    "parse",
    "parse_file",
    "parser",
    "AnyTalonRule",
    "ListValue",
    "find_command",
    "match",
    "VERSION",
]

################################################################################
# Properties for backwards compatibility
################################################################################


def _TalonMatch_key(self: TalonMatch) -> TalonIdentifier:
    warnings.warn(
        "'key' was deprecated in 1004.3.1.0; use 'left'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.left


setattr(
    TalonMatch,
    "key",
    property(_TalonMatch_key),
)


def _TalonMatch_modifier(self: TalonMatch) -> Sequence[TalonMatchModifier]:
    warnings.warn(
        "'modifier' was deprecated in 1004.3.1.0; use 'modifiers'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.modifiers


setattr(
    TalonMatch,
    "modifier",
    property(_TalonMatch_modifier),
)


def _TalonMatch_pattern(self: TalonMatch) -> TalonImplicitString:
    warnings.warn(
        "'pattern' was deprecated in 1004.3.1.0; use 'right'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.right


setattr(
    TalonMatch,
    "pattern",
    property(_TalonMatch_pattern),
)


def _TalonCommandDeclaration_rule(self: TalonCommandDeclaration) -> TalonRule:
    warnings.warn(
        "'rule' was deprecated in 1004.3.1.0; use 'left'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.left


setattr(
    TalonCommandDeclaration,
    "rule",
    property(_TalonCommandDeclaration_rule),
)


def _TalonCommandDeclaration_script(self: TalonCommandDeclaration) -> TalonBlock:
    warnings.warn(
        "'script' was deprecated in 1004.3.1.0; use 'right'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.right


setattr(
    TalonCommandDeclaration,
    "script",
    property(_TalonCommandDeclaration_script),
)


def _TalonKeyBindingDeclaration_key(self: TalonKeyBindingDeclaration) -> TalonKeyAction:
    warnings.warn(
        "'key' was deprecated in 1004.3.1.0; use 'left'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.left


setattr(
    TalonKeyBindingDeclaration,
    "key",
    property(_TalonKeyBindingDeclaration_key),
)


def _TalonKeyBindingDeclaration_script(self: TalonKeyBindingDeclaration) -> TalonBlock:
    warnings.warn(
        "'script' was deprecated in 1004.3.1.0; use 'right'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.right


setattr(
    TalonKeyBindingDeclaration,
    "script",
    property(_TalonKeyBindingDeclaration_script),
)


def _TalonTagImportDeclaration_tag(self: TalonTagImportDeclaration) -> TalonIdentifier:
    warnings.warn(
        "'tag' was deprecated in 1004.3.1.0; use 'right'",
        DeprecationWarning,
        stacklevel=2,
    )
    return self.right


setattr(
    TalonTagImportDeclaration,
    "tag",
    property(_TalonTagImportDeclaration_tag),
)

################################################################################
# Method to test if a matches block is explicit
################################################################################


def _TalonMatches_is_explicit(self: TalonMatches) -> bool:
    return self.text == "-" or self.text.endswith("\n-")


setattr(TalonMatches, "is_explicit", _TalonMatches_is_explicit)


################################################################################
# Redefine __init__ to merge comments into block
################################################################################


def _TalonBlock_with_comments(
    self: TalonBlock, comments: Optional[Sequence[TalonComment]]
) -> TalonBlock:
    return TalonBlock(
        text=self.text,
        type_name=self.type_name,
        start_position=self.start_position,
        end_position=self.end_position,
        children=[*(comments or ()), *self.children],
    )


def _TalonCommandDeclaration___init__(
    self: TalonCommandDeclaration,
    text: str,
    type_name: NodeTypeName,
    start_position: Point,
    end_position: Point,
    children: Optional[Sequence[TalonComment]],
    left: TalonRule,
    right: TalonBlock,
) -> None:
    self.text = text
    self.type_name = type_name
    self.start_position = start_position
    self.end_position = end_position
    self.children = None
    self.left = left
    self.right = _TalonBlock_with_comments(right, children)


setattr(TalonCommandDeclaration, "__init__", _TalonCommandDeclaration___init__)


def _TalonKeyBindingDeclaration___init__(
    self: TalonKeyBindingDeclaration,
    text: str,
    type_name: NodeTypeName,
    start_position: Point,
    end_position: Point,
    children: Optional[Sequence[TalonComment]],
    left: TalonKeyAction,
    right: TalonBlock,
) -> None:
    self.text = text
    self.type_name = type_name
    self.start_position = start_position
    self.end_position = end_position
    self.children = None
    self.left = left
    self.right = _TalonBlock_with_comments(right, children)


setattr(TalonKeyBindingDeclaration, "__init__", _TalonKeyBindingDeclaration___init__)


def _TalonSettingsDeclaration___init__(
    self: TalonSettingsDeclaration,
    text: str,
    type_name: NodeTypeName,
    start_position: Point,
    end_position: Point,
    children: Optional[Sequence[TalonComment]],
    right: TalonBlock,
) -> None:
    self.text = text
    self.type_name = type_name
    self.start_position = start_position
    self.end_position = end_position
    self.children = None
    self.right = _TalonBlock_with_comments(right, children)


setattr(TalonSettingsDeclaration, "__init__", _TalonSettingsDeclaration___init__)


################################################################################
# Method to test if a declaration is short
################################################################################


def _TalonBlock_is_short(self: TalonBlock) -> bool:
    return len(self.children) <= 1


setattr(TalonBlock, "is_short", _TalonBlock_is_short)


def _TalonCommandDeclaration_is_short(self: TalonCommandDeclaration) -> bool:
    return self.right.is_short()


setattr(TalonCommandDeclaration, "is_short", _TalonCommandDeclaration_is_short)


def _TalonSettingsDeclaration_is_short(self: TalonSettingsDeclaration) -> bool:
    return self.right.is_short()


setattr(TalonSettingsDeclaration, "is_short", _TalonSettingsDeclaration_is_short)


def _TalonKeyBindingDeclaration_is_short(self: TalonKeyBindingDeclaration) -> bool:
    return self.right.is_short()


setattr(TalonKeyBindingDeclaration, "is_short", _TalonKeyBindingDeclaration_is_short)


################################################################################
# Method to test if a node is extra
################################################################################


def _is_extra(self: Node) -> bool:
    """
    Check if a node is an extra node.
    """
    return isinstance(self, (TalonComment,))


setattr(Node, "is_extra", _is_extra)


################################################################################
# Method to get only non-extra child
################################################################################


def _get_child_fail_msg(self: Node, children: Iterable[Node] = []) -> str:
    return "\n".join(
        [
            f"Node '{self.type_name}' should have exactly one non-comment child, found:",
            *[f"- {child}" for child in children],
            f"in {self}",
        ]
    )


def _get_child(self: Node) -> Node:
    """
    Get the only non-extra child of a branch node.
    """
    assert isinstance(self, Branch)
    assert self.children is not None, _get_child_fail_msg(self)
    if isinstance(self.children, Node):
        assert not _is_extra(self.children), _get_child_fail_msg(
            self, children=[self.children]
        )
        return self.children
    else:
        for i in range(0, len(self.children)):
            if _is_extra(self.children[i]):
                continue
            assert all(
                _is_extra(child) for child in self.children[i + 1 :]
            ), _get_child_fail_msg(self, children=self.children)
            return self.children[i]
        raise AssertionError(_get_child_fail_msg(self, children=self.children))


setattr(TalonOptional, "get_child", _get_child)
setattr(TalonParenthesizedRule, "get_child", _get_child)
setattr(TalonRepeat, "get_child", _get_child)
setattr(TalonRepeat1, "get_child", _get_child)
setattr(TalonParenthesizedExpression, "get_child", _get_child)
setattr(TalonInterpolation, "get_child", _get_child)

################################################################################
# Method to get docstrings
##############################################_removeprefix##################################


def _TalonComment_get_docstring(self: TalonComment) -> Optional[str]:
    if self.text.startswith("###"):
        return _removeprefix(self.text, "###").strip()
    return None


setattr(TalonComment, "get_docstring", _TalonComment_get_docstring)


def _TalonSourceFile_get_docstring(self: TalonSourceFile) -> Optional[str]:
    docstrings: List[str] = []
    for comment in self.children:
        if isinstance(comment, TalonComment):
            docstring = _TalonComment_get_docstring(comment)
            if docstring:
                docstrings.append(docstring)
        else:
            break
    return "\n".join(docstrings) if docstrings else None


setattr(TalonSourceFile, "get_docstring", _TalonSourceFile_get_docstring)


def _TalonCommandDeclaration_get_docstring(
    self: TalonCommandDeclaration,
) -> Optional[str]:
    docstrings: List[str] = []
    for comment in self.right.children:
        if isinstance(comment, TalonComment):
            docstring = _TalonComment_get_docstring(comment)
            if docstring:
                docstrings.append(docstring)
    return "\n".join(docstrings) if docstrings else None


setattr(
    TalonCommandDeclaration, "get_docstring", _TalonCommandDeclaration_get_docstring
)


################################################################################
# Pickle Compatibility
################################################################################


def _fields(obj: Type[Any]) -> Tuple[Any, ...]:
    assert is_dataclass(obj)
    return tuple(getattr(obj, field.name) for field in fields(obj))


def _make_TalonAction(*args: Any) -> TalonAction:
    return TalonAction(*args)


setattr(
    TalonAction,
    "__reduce__",
    lambda self: (_make_TalonAction, _fields(self)),
)


def _make_TalonArgumentList(*args: Any) -> TalonArgumentList:
    return TalonArgumentList(*args)


setattr(
    TalonArgumentList,
    "__reduce__",
    lambda self: (_make_TalonArgumentList, _fields(self)),
)


def _make_TalonAssignmentStatement(*args: Any) -> TalonAssignmentStatement:
    return TalonAssignmentStatement(*args)


setattr(
    TalonAssignmentStatement,
    "__reduce__",
    lambda self: (_make_TalonAssignmentStatement, _fields(self)),
)


def _make_TalonBinaryOperator(*args: Any) -> TalonBinaryOperator:
    return TalonBinaryOperator(*args)


setattr(
    TalonBinaryOperator,
    "__reduce__",
    lambda self: (_make_TalonBinaryOperator, _fields(self)),
)


def _make_TalonBlock(*args: Any) -> TalonBlock:
    return TalonBlock(*args)


setattr(
    TalonBlock,
    "__reduce__",
    lambda self: (_make_TalonBlock, _fields(self)),
)


def _make_TalonCapture(*args: Any) -> TalonCapture:
    return TalonCapture(*args)


setattr(
    TalonCapture,
    "__reduce__",
    lambda self: (_make_TalonCapture, _fields(self)),
)


def _make_TalonChoice(*args: Any) -> TalonChoice:
    return TalonChoice(*args)


setattr(
    TalonChoice,
    "__reduce__",
    lambda self: (_make_TalonChoice, _fields(self)),
)


def _make_TalonCommandDeclaration(*args: Any) -> TalonCommandDeclaration:
    return TalonCommandDeclaration(*args)


setattr(
    TalonCommandDeclaration,
    "__reduce__",
    lambda self: (_make_TalonCommandDeclaration, _fields(self)),
)


def _make_TalonComment(*args: Any) -> TalonComment:
    return TalonComment(*args)


setattr(
    TalonComment,
    "__reduce__",
    lambda self: (_make_TalonComment, _fields(self)),
)


def _make_TalonDeclarations(*args: Any) -> TalonDeclarations:
    return TalonDeclarations(*args)


setattr(
    TalonDeclarations,
    "__reduce__",
    lambda self: (_make_TalonDeclarations, _fields(self)),
)


def _make_TalonDeclaration(*args: Any) -> TalonDeclaration:
    return TalonDeclaration(*args)


setattr(
    TalonDeclaration,
    "__reduce__",
    lambda self: (_make_TalonDeclaration, _fields(self)),
)


def _make_TalonEndAnchor(*args: Any) -> TalonEndAnchor:
    return TalonEndAnchor(*args)


setattr(
    TalonEndAnchor,
    "__reduce__",
    lambda self: (_make_TalonEndAnchor, _fields(self)),
)


def _make_TalonError(*args: Any) -> TalonError:
    return TalonError(*args)


setattr(
    TalonError,
    "__reduce__",
    lambda self: (_make_TalonError, _fields(self)),
)


def _make_TalonExpression(*args: Any) -> TalonExpression:
    return TalonExpression(*args)


setattr(
    TalonExpression,
    "__reduce__",
    lambda self: (_make_TalonExpression, _fields(self)),
)


def _make_TalonExpressionStatement(*args: Any) -> TalonExpressionStatement:
    return TalonExpressionStatement(*args)


setattr(
    TalonExpressionStatement,
    "__reduce__",
    lambda self: (_make_TalonExpressionStatement, _fields(self)),
)


def _make_TalonFloat(*args: Any) -> TalonFloat:
    return TalonFloat(*args)


setattr(
    TalonFloat,
    "__reduce__",
    lambda self: (_make_TalonFloat, _fields(self)),
)


def _make_TalonIdentifier(*args: Any) -> TalonIdentifier:
    return TalonIdentifier(*args)


setattr(
    TalonIdentifier,
    "__reduce__",
    lambda self: (_make_TalonIdentifier, _fields(self)),
)


def _make_TalonImplicitString(*args: Any) -> TalonImplicitString:
    return TalonImplicitString(*args)


setattr(
    TalonImplicitString,
    "__reduce__",
    lambda self: (_make_TalonImplicitString, _fields(self)),
)


def _make_TalonInteger(*args: Any) -> TalonInteger:
    return TalonInteger(*args)


setattr(
    TalonInteger,
    "__reduce__",
    lambda self: (_make_TalonInteger, _fields(self)),
)


def _make_TalonInterpolation(*args: Any) -> TalonInterpolation:
    return TalonInterpolation(*args)


setattr(
    TalonInterpolation,
    "__reduce__",
    lambda self: (_make_TalonInterpolation, _fields(self)),
)


def _make_TalonKeyAction(*args: Any) -> TalonKeyAction:
    return TalonKeyAction(*args)


setattr(
    TalonKeyAction,
    "__reduce__",
    lambda self: (_make_TalonKeyAction, _fields(self)),
)


def _make_TalonKeyBindingDeclaration(*args: Any) -> TalonKeyBindingDeclaration:
    return TalonKeyBindingDeclaration(*args)


setattr(
    TalonKeyBindingDeclaration,
    "__reduce__",
    lambda self: (_make_TalonKeyBindingDeclaration, _fields(self)),
)


def _make_TalonList(*args: Any) -> TalonList:
    return TalonList(*args)


setattr(
    TalonList,
    "__reduce__",
    lambda self: (_make_TalonList, _fields(self)),
)


def _make_TalonMatch(*args: Any) -> TalonMatch:
    return TalonMatch(*args)


setattr(
    TalonMatch,
    "__reduce__",
    lambda self: (_make_TalonMatch, _fields(self)),
)


def _make_TalonMatches(*args: Any) -> TalonMatches:
    return TalonMatches(*args)


setattr(
    TalonMatches,
    "__reduce__",
    lambda self: (_make_TalonMatches, _fields(self)),
)


def _make_TalonMatchModifier(*args: Any) -> TalonMatchModifier:
    return TalonMatchModifier(*args)


setattr(
    TalonMatchModifier,
    "__reduce__",
    lambda self: (_make_TalonMatchModifier, _fields(self)),
)


def _make_TalonNumber(*args: Any) -> TalonNumber:
    return TalonNumber(*args)


setattr(
    TalonNumber,
    "__reduce__",
    lambda self: (_make_TalonNumber, _fields(self)),
)


def _make_TalonOperator(*args: Any) -> TalonOperator:
    return TalonOperator(*args)


setattr(
    TalonOperator,
    "__reduce__",
    lambda self: (_make_TalonOperator, _fields(self)),
)


def _make_TalonOptional(*args: Any) -> TalonOptional:
    return TalonOptional(*args)


setattr(
    TalonOptional,
    "__reduce__",
    lambda self: (_make_TalonOptional, _fields(self)),
)


def _make_TalonParenthesizedExpression(*args: Any) -> TalonParenthesizedExpression:
    return TalonParenthesizedExpression(*args)


setattr(
    TalonParenthesizedExpression,
    "__reduce__",
    lambda self: (_make_TalonParenthesizedExpression, _fields(self)),
)


def _make_TalonParenthesizedRule(*args: Any) -> TalonParenthesizedRule:
    return TalonParenthesizedRule(*args)


setattr(
    TalonParenthesizedRule,
    "__reduce__",
    lambda self: (_make_TalonParenthesizedRule, _fields(self)),
)


def _make_TalonRepeat(*args: Any) -> TalonRepeat:
    return TalonRepeat(*args)


setattr(
    TalonRepeat,
    "__reduce__",
    lambda self: (_make_TalonRepeat, _fields(self)),
)


def _make_TalonRepeat1(*args: Any) -> TalonRepeat1:
    return TalonRepeat1(*args)


setattr(
    TalonRepeat1,
    "__reduce__",
    lambda self: (_make_TalonRepeat1, _fields(self)),
)


def _make_TalonRule(*args: Any) -> TalonRule:
    return TalonRule(*args)


setattr(
    TalonRule,
    "__reduce__",
    lambda self: (_make_TalonRule, _fields(self)),
)


def _make_TalonSeq(*args: Any) -> TalonSeq:
    return TalonSeq(*args)


setattr(
    TalonSeq,
    "__reduce__",
    lambda self: (_make_TalonSeq, _fields(self)),
)


def _make_TalonSettingsDeclaration(*args: Any) -> TalonSettingsDeclaration:
    return TalonSettingsDeclaration(*args)


setattr(
    TalonSettingsDeclaration,
    "__reduce__",
    lambda self: (_make_TalonSettingsDeclaration, _fields(self)),
)


def _make_TalonSleepAction(*args: Any) -> TalonSleepAction:
    return TalonSleepAction(*args)


setattr(
    TalonSleepAction,
    "__reduce__",
    lambda self: (_make_TalonSleepAction, _fields(self)),
)


def _make_TalonSourceFile(*args: Any) -> TalonSourceFile:
    return TalonSourceFile(*args)


setattr(
    TalonSourceFile,
    "__reduce__",
    lambda self: (_make_TalonSourceFile, _fields(self)),
)


def _make_TalonStartAnchor(*args: Any) -> TalonStartAnchor:
    return TalonStartAnchor(*args)


setattr(
    TalonStartAnchor,
    "__reduce__",
    lambda self: (_make_TalonStartAnchor, _fields(self)),
)


def _make_TalonStatement(*args: Any) -> TalonStatement:
    return TalonStatement(*args)


setattr(
    TalonStatement,
    "__reduce__",
    lambda self: (_make_TalonStatement, _fields(self)),
)


def _make_TalonString(*args: Any) -> TalonString:
    return TalonString(*args)


setattr(
    TalonString,
    "__reduce__",
    lambda self: (_make_TalonString, _fields(self)),
)


def _make_TalonStringContent(*args: Any) -> TalonStringContent:
    return TalonStringContent(*args)


setattr(
    TalonStringContent,
    "__reduce__",
    lambda self: (_make_TalonStringContent, _fields(self)),
)


def _make_TalonStringEscapeSequence(*args: Any) -> TalonStringEscapeSequence:
    return TalonStringEscapeSequence(*args)


setattr(
    TalonStringEscapeSequence,
    "__reduce__",
    lambda self: (_make_TalonStringEscapeSequence, _fields(self)),
)


def _make_TalonTagImportDeclaration(*args: Any) -> TalonTagImportDeclaration:
    return TalonTagImportDeclaration(*args)


setattr(
    TalonTagImportDeclaration,
    "__reduce__",
    lambda self: (_make_TalonTagImportDeclaration, _fields(self)),
)


def _make_TalonUnaryOperator(*args: Any) -> TalonUnaryOperator:
    return TalonUnaryOperator(*args)


setattr(
    TalonUnaryOperator,
    "__reduce__",
    lambda self: (_make_TalonUnaryOperator, _fields(self)),
)


def _make_TalonVariable(*args: Any) -> TalonVariable:
    return TalonVariable(*args)


setattr(
    TalonVariable,
    "__reduce__",
    lambda self: (_make_TalonVariable, _fields(self)),
)


def _make_TalonWord(*args: Any) -> TalonWord:
    return TalonWord(*args)


setattr(
    TalonWord,
    "__reduce__",
    lambda self: (_make_TalonWord, _fields(self)),
)



from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            name="tree_sitter_talon._internal.binding",
            sources=[
                "src/tree_sitter_talon/_internal/binding.c",
                "src/tree_sitter_talon/_tree_sitter_talon/src/parser.c",
                "src/tree_sitter_talon/_tree_sitter_talon/src/scanner.cc",
            ],
            include_dirs=[
                "src/tree_sitter_talon/_tree_sitter_talon/src",
            ],
        )
    ],
)



from pathlib import Path
from typing import List, Optional, Sequence, Union

import importlib_resources
import tree_sitter
import tree_sitter_type_provider

from .binding import _tree_sitter_talon_id

################################################################################
# Export List
################################################################################

__all__: List[str] = [
    "TalonLanguage",
    "TreeSitterTalon",
]


################################################################################
# Core Classes
################################################################################


class TalonLanguage(tree_sitter.Language):  # type: ignore
    def __init__(self) -> None:
        super().__init__(_tree_sitter_talon_id(), "talon")


class TreeSitterTalon(tree_sitter_type_provider.TreeSitterTypeProvider):  # type: ignore
    _library_path: Optional[str] = None

    _language: Optional[tree_sitter.Language] = None

    _parser: Optional[tree_sitter.Parser] = None

    @property
    def _node_types(
        self,
    ) -> Sequence[tree_sitter_type_provider.NodeType]:
        _node_types_text = (
            importlib_resources.files("tree_sitter_talon")
            .joinpath("_tree_sitter_talon", "src", "node-types.json")
            .read_text(encoding="utf-8")
        )
        return tree_sitter_type_provider.NodeType.schema().loads(  # type: ignore
            _node_types_text, many=True
        )

    @property
    def language(self) -> tree_sitter.Language:
        if not self._language:
            self._language = TalonLanguage()
        return self._language

    @property
    def parser(self) -> tree_sitter.Parser:
        if not self._parser:
            self._parser = tree_sitter.Parser()
            self._parser.set_language(self.language)
        return self._parser

    def __init__(self) -> None:
        # Conversion from tree-sitter names to Python names
        def as_class_name(node_type_name: str) -> str:
            buffer = ["Talon"]
            for part in node_type_name.split("_"):
                buffer.append(part.capitalize())
            return "".join(buffer)

        # Initialize module
        super().__init__(
            "tree_sitter_talon",
            self._node_types,
            as_class_name=as_class_name,
            extra=["comment"],
        )
        self.NodeTypeError = tree_sitter_type_provider.NodeTypeError
        self.NodeTypeName = tree_sitter_type_provider.NodeTypeName
        self.NodeFieldName = tree_sitter_type_provider.NodeFieldName
        self.Node = tree_sitter_type_provider.Node
        self.Leaf = tree_sitter_type_provider.Leaf
        self.Branch = tree_sitter_type_provider.Branch
        self.Point = tree_sitter_type_provider.Point
        self.ParseError = tree_sitter_type_provider.ParseError

        ################################################################################
        # Export List
        ################################################################################

        self.__all__: List[str] = [
            "parser",
            "language",
            "parse",
            "parse_file",
            "from_tree_sitter",
            "NodeTypeName",
            "NodeFieldName",
            "NodeTypeError",
            "Point",
            "Node",
            "Leaf",
            "Branch",
            "ParseError",
            "TalonSourceFile",
            "TalonComment",
            "TalonError",
            "TalonMatches",
            "TalonMatch",
            "TalonMatchModifier",
            "TalonDeclarations",
            "TalonDeclaration",
            "TalonCommandDeclaration",
            "TalonKeyBindingDeclaration",
            "TalonSettingsDeclaration",
            "TalonTagImportDeclaration",
            "TalonCapture",
            "TalonChoice",
            "TalonEndAnchor",
            "TalonList",
            "TalonOptional",
            "TalonParenthesizedRule",
            "TalonRepeat",
            "TalonRepeat1",
            "TalonRule",
            "TalonSeq",
            "TalonStartAnchor",
            "TalonWord",
            "TalonStatement",
            "TalonAssignmentStatement",
            "TalonExpressionStatement",
            "TalonBlock",
            "TalonExpression",
            "TalonAction",
            "TalonArgumentList",
            "TalonUnaryOperator",
            "TalonBinaryOperator",
            "TalonKeyAction",
            "TalonParenthesizedExpression",
            "TalonSleepAction",
            "TalonVariable",
            "TalonIdentifier",
            "TalonOperator",
            "TalonImplicitString",
            "TalonInterpolation",
            "TalonString",
            "TalonStringContent",
            "TalonStringEscapeSequence",
            "TalonNumber",
            "TalonFloat",
            "TalonInteger",
        ]

    def parse(
        self,
        contents: Union[str, bytes],
        *,
        encoding: str = "utf-8",
        filename: Optional[str] = None,
        raise_parse_error: bool = False,
    ) -> Optional[
        Union[
            tree_sitter_type_provider.Node,
            Sequence[tree_sitter_type_provider.Node],
        ]
    ]:
        tree = self._parse(contents, encoding=encoding)
        return self.from_tree_sitter(
            tree.root_node,
            encoding=encoding,
            filename=filename,
            raise_parse_error=raise_parse_error,
        )

    def parse_file(
        self,
        path: Union[str, Path],
        *,
        encoding: str = "utf-8",
        raise_parse_error: bool = False,
    ) -> Optional[
        Union[
            tree_sitter_type_provider.Node,
            Sequence[tree_sitter_type_provider.Node],
        ]
    ]:
        tree = self._parse_file(path, encoding=encoding)
        return self.from_tree_sitter(
            tree.root_node,
            encoding=encoding,
            filename=str(path),
            raise_parse_error=raise_parse_error,
        )

    def _parse(
        self,
        contents: Union[str, bytes],
        *,
        encoding: str = "utf-8",
    ) -> tree_sitter.Tree:
        if isinstance(contents, str):
            contents = bytes(contents, encoding)
        return self.parser.parse(contents)

    def _parse_file(
        self,
        path: Union[str, Path],
        *,
        encoding: str = "utf-8",
    ) -> tree_sitter.Tree:
        if not isinstance(path, Path):
            path = Path(path)
        contents = path.read_bytes()
        return self._parse(contents, encoding=encoding)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, List, Optional, Sequence, Tuple, Union

import tree_sitter
from dataclasses_json import DataClassJsonMixin, config

from .match import AnyTalonRule, ListValue

################################################################################
# Export List
################################################################################

__all__: List[str] = [
    "parser",
    "language",
    "parse",
    "parse_file",
    "from_tree_sitter",
    "NodeTypeName",
    "NodeFieldName",
    "NodeTypeError",
    "Point",
    "Node",
    "Leaf",
    "Branch",
    "ParseError",
    "TalonSourceFile",
    "TalonComment",
    "TalonError",
    "TalonMatches",
    "TalonMatch",
    "TalonMatchModifier",
    "TalonDeclarations",
    "TalonDeclaration",
    "TalonCommandDeclaration",
    "TalonKeyBindingDeclaration",
    "TalonSettingsDeclaration",
    "TalonTagImportDeclaration",
    "TalonCapture",
    "TalonChoice",
    "TalonEndAnchor",
    "TalonList",
    "TalonOptional",
    "TalonParenthesizedRule",
    "TalonRepeat",
    "TalonRepeat1",
    "TalonRule",
    "TalonSeq",
    "TalonStartAnchor",
    "TalonWord",
    "TalonStatement",
    "TalonAssignmentStatement",
    "TalonExpressionStatement",
    "TalonBlock",
    "TalonExpression",
    "TalonAction",
    "TalonArgumentList",
    "TalonUnaryOperator",
    "TalonBinaryOperator",
    "TalonKeyAction",
    "TalonParenthesizedExpression",
    "TalonSleepAction",
    "TalonVariable",
    "TalonIdentifier",
    "TalonOperator",
    "TalonImplicitString",
    "TalonInterpolation",
    "TalonString",
    "TalonStringContent",
    "TalonStringEscapeSequence",
    "TalonNumber",
    "TalonFloat",
    "TalonInteger",
]

################################################################################
# Extended node types (from tree-sitter-type-provider)
################################################################################

class NodeTypeError(Exception):
    pass

@dataclass
class Point:
    line: int
    column: int

    @staticmethod
    def from_tree_sitter(tspoint: Tuple[int, int]) -> "Point":
        return Point(line=tspoint[0], column=tspoint[1])

NodeTypeName = str

NodeFieldName = str

@dataclass
class Node(DataClassJsonMixin):
    text: str
    type_name: NodeTypeName = field(metadata=config(field_name="type"))
    start_position: Point
    end_position: Point

    def is_extra(self) -> bool: ...
    def is_equivalent(self, other: "Node") -> bool: ...
    def assert_equivalent(self, other: "Node") -> None: ...

@dataclass
class Leaf(Node):
    pass

@dataclass
class Branch(Node):
    children: Union[None, Node, Sequence[Node]]

@dataclass
class ParseError(Exception, Branch):
    children: List[Node]
    contents: Optional[str] = None
    filename: Optional[str] = None

parser: tree_sitter.Parser

language: tree_sitter.Language

def parse(
    contents: Union[str, bytes],
    *,
    encoding: str = "utf-8",
    filename: Optional[str] = None,
    raise_parse_error: bool = False,
) -> Node: ...
def parse_file(
    path: Union[str, Path],
    *,
    encoding: str = "utf-8",
    raise_parse_error: bool = False,
) -> Node: ...
def from_tree_sitter(
    tsvalue: Union[tree_sitter.Tree, tree_sitter.Node, tree_sitter.TreeCursor],
    *,
    encoding: str = "utf-8",
    filename: Optional[str] = None,
    raise_parse_error: bool = False,
) -> Node: ...

# AST node classes.

@dataclass
class TalonSourceFile(Branch):
    children: Sequence[Union[TalonDeclarations, TalonMatches, TalonComment]]
    def get_docstring(self) -> Optional[str]: ...
    def find_command(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> Optional["TalonCommandDeclaration"]: ...

@dataclass
class TalonComment(Leaf):
    def get_docstring(self) -> Optional[str]: ...

@dataclass
class TalonError(Exception, Branch):
    children: Sequence[Node]
    contents: Optional[str] = None
    filename: Optional[str] = None

# Matches.

@dataclass
class TalonMatches(Branch):
    children: Sequence[Union[TalonMatch, TalonComment]]

    def is_explicit(self) -> bool: ...

@dataclass
class TalonMatch(Branch):
    children: Sequence[TalonComment]
    left: TalonIdentifier
    modifiers: Sequence[TalonMatchModifier]
    right: TalonImplicitString

    @property
    def key(self) -> TalonIdentifier: ...
    @property
    def modifier(self) -> Sequence[TalonMatchModifier]: ...
    @property
    def pattern(self) -> TalonImplicitString: ...

@dataclass
class TalonMatchModifier(Leaf):
    pass

# Declarations.

@dataclass
class TalonDeclarations(Branch):
    children: Sequence[TalonDeclaration]

@dataclass
class TalonDeclaration(Node):
    pass

@dataclass(init=False)
class TalonCommandDeclaration(Branch, TalonDeclaration):
    children: None
    left: TalonRule
    right: TalonBlock

    def __init__(
        self,
        text: str,
        type_name: NodeTypeName,
        start_position: Point,
        end_position: Point,
        children: Optional[Sequence[TalonComment]],
        left: TalonRule,
        right: TalonBlock,
    ) -> None: ...
    @property
    def rule(self) -> TalonRule: ...
    @property
    def script(self) -> TalonBlock: ...
    def get_docstring(self) -> Optional[str]: ...
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...
    def is_short(self) -> bool: ...

@dataclass(init=False)
class TalonKeyBindingDeclaration(Branch, TalonDeclaration):
    children: None
    left: TalonKeyAction
    right: TalonBlock

    def __init__(
        self,
        text: str,
        type_name: NodeTypeName,
        start_position: Point,
        end_position: Point,
        children: Optional[Sequence[TalonComment]],
        left: TalonKeyAction,
        right: TalonBlock,
    ) -> None: ...
    @property
    def key(self) -> TalonKeyAction: ...
    @property
    def script(self) -> TalonBlock: ...
    def is_short(self) -> bool: ...

@dataclass(init=False)
class TalonSettingsDeclaration(Branch, TalonDeclaration):
    children: None
    right: TalonBlock

    def __init__(
        self,
        text: str,
        type_name: NodeTypeName,
        start_position: Point,
        end_position: Point,
        children: Optional[Sequence[TalonComment]],
        right: TalonBlock,
    ) -> None: ...
    def is_short(self) -> bool: ...

@dataclass
class TalonTagImportDeclaration(Branch, TalonDeclaration):
    children: Sequence[TalonComment]
    right: TalonIdentifier

    @property
    def tag(self) -> TalonIdentifier: ...

# Rules.

@dataclass
class TalonCapture(Branch):
    children: Sequence[TalonComment]
    capture_name: TalonIdentifier
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonChoice(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonEndAnchor,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonSeq,
            TalonStartAnchor,
            TalonWord,
            TalonComment,
        ]
    ]
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonEndAnchor(Leaf):
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonList(Branch):
    children: Sequence[TalonComment]
    list_name: TalonIdentifier
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonOptional(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonChoice,
            TalonEndAnchor,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonSeq,
            TalonStartAnchor,
            TalonWord,
            TalonComment,
        ]
    ]
    def get_child(
        self,
    ) -> Union[
        TalonCapture,
        TalonChoice,
        TalonEndAnchor,
        TalonList,
        TalonOptional,
        TalonParenthesizedRule,
        TalonRepeat,
        TalonRepeat1,
        TalonSeq,
        TalonStartAnchor,
        TalonWord,
    ]: ...
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonParenthesizedRule(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonChoice,
            TalonEndAnchor,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonSeq,
            TalonStartAnchor,
            TalonWord,
            TalonComment,
        ]
    ]
    def get_child(
        self,
    ) -> Union[
        TalonCapture,
        TalonChoice,
        TalonEndAnchor,
        TalonList,
        TalonOptional,
        TalonParenthesizedRule,
        TalonRepeat,
        TalonRepeat1,
        TalonSeq,
        TalonStartAnchor,
        TalonWord,
    ]: ...
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonRepeat(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonWord,
            TalonComment,
        ]
    ]
    def get_child(
        self,
    ) -> Union[
        TalonCapture,
        TalonList,
        TalonOptional,
        TalonParenthesizedRule,
        TalonRepeat,
        TalonRepeat1,
        TalonWord,
    ]: ...
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonRepeat1(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonWord,
            TalonComment,
        ]
    ]
    def get_child(
        self,
    ) -> Union[
        TalonCapture,
        TalonList,
        TalonOptional,
        TalonParenthesizedRule,
        TalonRepeat,
        TalonRepeat1,
        TalonWord,
    ]: ...
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonRule(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonChoice,
            TalonEndAnchor,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonSeq,
            TalonStartAnchor,
            TalonWord,
            TalonComment,
        ]
    ]
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonSeq(Branch):
    children: Sequence[
        Union[
            TalonCapture,
            TalonList,
            TalonOptional,
            TalonParenthesizedRule,
            TalonRepeat,
            TalonRepeat1,
            TalonWord,
            TalonComment,
        ]
    ]
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonStartAnchor(Leaf):
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

@dataclass
class TalonWord(Leaf):
    def match(
        self,
        text: Sequence[str],
        *,
        fullmatch: bool = False,
        get_capture: Optional[Callable[[str], Optional[AnyTalonRule]]] = None,
        get_list: Optional[Callable[[str], Optional[ListValue]]] = None,
    ) -> bool: ...

# Statements.

@dataclass
class TalonStatement(Node):
    pass

@dataclass
class TalonAssignmentStatement(Branch, TalonStatement):
    children: Sequence[TalonComment]
    left: TalonIdentifier
    right: TalonExpression

@dataclass
class TalonExpressionStatement(Branch, TalonStatement):
    children: Sequence[TalonComment]
    expression: TalonExpression

@dataclass
class TalonBlock(Branch):
    children: Sequence[Union[TalonStatement, TalonComment]]

    def is_short(self) -> bool: ...

# Expressions.

@dataclass
class TalonExpression(Node):
    pass

@dataclass
class TalonAction(Branch, TalonExpression):
    children: Sequence[TalonComment]
    action_name: TalonIdentifier
    arguments: TalonArgumentList

@dataclass
class TalonArgumentList(Branch):
    children: Sequence[Union[TalonExpression, TalonComment]]

@dataclass
class TalonUnaryOperator(Branch, TalonExpression):
    children: Sequence[TalonComment]
    operator: TalonOperator
    right: TalonExpression

@dataclass
class TalonBinaryOperator(Branch, TalonExpression):
    children: Sequence[TalonComment]
    left: TalonExpression
    operator: TalonOperator
    right: TalonExpression

@dataclass
class TalonKeyAction(Branch, TalonExpression):
    children: Sequence[TalonComment]
    arguments: TalonImplicitString

@dataclass
class TalonParenthesizedExpression(Branch, TalonExpression):
    children: Sequence[Union[TalonExpression, TalonComment]]

    def get_child(self) -> TalonExpression: ...

@dataclass
class TalonSleepAction(Branch, TalonExpression):
    children: Sequence[TalonComment]
    arguments: TalonImplicitString

@dataclass
class TalonVariable(Branch, TalonExpression):
    children: Sequence[TalonComment]
    variable_name: TalonIdentifier

# Identifiers.

@dataclass
class TalonIdentifier(Leaf):
    pass

@dataclass
class TalonOperator(Leaf):
    pass

# Strings.

@dataclass
class TalonImplicitString(Leaf):
    pass

@dataclass
class TalonInterpolation(Branch):
    children: Sequence[Union[TalonExpression, TalonComment]]

    def get_child(self) -> TalonExpression: ...

@dataclass
class TalonString(Branch):
    children: Sequence[
        Union[
            TalonInterpolation,
            TalonStringContent,
            TalonStringEscapeSequence,
            TalonComment,
        ]
    ]

@dataclass
class TalonStringContent(Leaf):
    pass

@dataclass
class TalonStringEscapeSequence(Leaf):
    pass

# Numbers.

@dataclass
class TalonNumber(Node):
    pass

@dataclass
class TalonFloat(Leaf, TalonNumber):
    pass

@dataclass
class TalonInteger(Leaf, TalonNumber):
    pass
